/*
** buffer.c - some buffer routines used by the fury
**
** Keith Zaback (krz@wcug.wwu.edu) Apr 98
*/

/* includes */
#include "ipxfury.h"
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

/* defines */
#define BUFFER_PARANOID		1

/* protos */
int			buffer_move(BUFFER *buf);


/*
**	BUFFER_GET_ALL -
*/
void buffer_get_all(BUFFER *buf, BYTE **ptr, int *len)
{

	/* return the ptr and len */
	*ptr = &buf->data[buf->read_index];
	*len = buf->write_index - buf->read_index;
	
	/* adjust the buffer to be empty */
	buf->read_index 	= 0;
	buf->write_index 	= 0;
	buf->dirty		= 0;
	
	/* done */
}



/*
**	BUFFER_CLEAR - cleans a buffer
*/
void buffer_clear(BUFFER *buf)
{
	/* set all the stuff to zero */
	buf->read_index 	= 0;
	buf->write_index 	= 0;
	buf->dirty		= 0;
	
	/* done */
}



/*
**	BUFFER_COMPACT
*/
int buffer_compact(BUFFER *buf)
{
	if( buf->read_index > 0 ) {	/* something to move */
		memcpy(buf->data, &buf->data[buf->read_index], (buf->write_index - buf->read_index));
		buf->write_index -= buf->read_index;	/* adjust ptrs */
		buf->read_index = 0;
		return(1);				/* gained some space */
	}
	
	/* nothing to do */
	return(0);
}




/*
**	BUFFER_INSERT()
*/
int buffer_insert(BUFFER *buf, unsigned char *data, int datalen)
{
	int		bytes_stored;

#if BUFFER_PARANOID == 1	
	/* make sure we have data storage */
	if( buf->data == NULL ) {
		errormsg(0, "buffer_insert", "buffer has no data");
		return(0);
	}
#endif
	
	/* do we have the space? */
	if( (buf->total_size - buf->write_index) < datalen ) {
		buffer_compact(buf);		/* slide the data */
	}
	
	/* can we fit it all? */
	if( (buf->total_size - buf->write_index) >= datalen ) 
		bytes_stored = datalen;
	else 		/* it can not fit! */
		return(0);
	
	/* copy the data */
	memcpy(&buf->data[buf->write_index], data, bytes_stored);
	
	/* move write index */
	buf->write_index += bytes_stored;

	/* set dirty bit */
	buf->dirty = 1;
			
	/* return bytes inserted */
	return(bytes_stored);

}




/*
**	BUFFER_GET()
*/
int buffer_get(BUFFER *buf, unsigned char *data, int datalen)
{
	int		bytes_recv;
	int		bytes_avail;

#if BUFFER_PARANOID == 1	
	/* do we have valid data */
	if( buf->data == NULL ) {
		errormsg(0, "buffer_get", "buffer has no data");
		return(0);
	}
#endif

	/* calc bytes in buffer */
	bytes_avail = buf->write_index - buf->read_index;
	
	/* can we give them all they want? */
	if( bytes_avail >= datalen )
		bytes_recv = datalen;		/* give them all */
	else
		bytes_recv = bytes_avail;	/* give them what we have */
		
	/* copy the data */
	memcpy(data, &buf->data[buf->read_index], bytes_recv);
	
	/* adjust the index */
	buf->read_index += bytes_recv;
	
	/* are read & write pointing at the same spot? */
	if( buf->read_index == buf->write_index ) {
		buf->read_index = 0;		/* reset them to 0 */
		buf->write_index = 0;
		buf->dirty = 0;		/* clear dirty bit */
	}
	
	/* return bytes read */	
	return(bytes_recv);

}




/*
**	BUFFER_INSERTB - insert byte
*/
int buffer_insertb(BUFFER *buf, unsigned char ch)
{

#if BUFFER_PARANOID == 1	
	/* do we have valid data */
	if( buf->data == NULL ) {
		errormsg(0, "buffer_insertb", "buffer has no data");
		return(0);
	}
#endif
	
	/* can this fit */
	if( buf->write_index < (buf->total_size-1) ) {
		buf->data[buf->write_index++] = ch;
		buf->dirty = 1;
		return(1);
	} else {		/* compact it, and try again */
		buffer_compact(buf);
		if( buf->write_index < (buf->total_size-1) ) {
			buf->data[buf->write_index++] = ch;
			buf->dirty = 1;
			return(1);
		}
	}
	
	/* failed */
	return(0);
}




/*
**	BUFFER_GETB
*/
int buffer_getb(BUFFER *buf, unsigned char *ch)
{

#if BUFFER_PARANOID == 1	
	/* do we have valid data */
	if( buf->data == NULL ) {
		errormsg(0, "buffer_insertb", "buffer has no data");
		return(0);
	}
#endif
	/* check that we have something */
	if( buf->read_index < buf->write_index ) {
		*ch = buf->data[buf->read_index++];
		if( buf->read_index == buf->write_index ) {
			buf->read_index = 0;		/* reset them */
			buf->write_index = 0;
			buf->dirty = 0;		/* empty */
		}
		return(1);
	} 
	
	/* failed */
	return(0);
}




/*
**	CREATE_BUFFER()
*/
int create_buffer(BUFFER *buf, int size)
{

	/* clean the struct */
	memset(buf, 0, sizeof(BUFFER));
	
	/* alloc the piece they want */
	buf->data = (unsigned char *)malloc(size);
	if( buf->data != NULL ) {
		memset(buf->data, 0, size);	/* clean it out */
		buf->total_size = size;		/* set size */
		/* read & write index are both zero from the memset */
		return(1);			/* success */
	} 
	
	/* failed */
	errormsg(0, "create_buffer", "could not allocate buffer (%d bytes)", size);
	return(0);
}



/*
**	FREE_BUFFER()
*/
void free_buffer(BUFFER *buf)
{
	/* free the buffer */
	if( buf->data != NULL )
		free(buf->data);
	
	/* clean the struct */
	memset(buf, 0, sizeof(BUFFER));
	
	/* done */
	return;
}



