/*
** fury.c - the wonky protocol used over the modem
**
** Keith Zaback (krz@wcug.wwu.edu) Apr 98
*/

/* user stuff */
#define SHOW_STATE_CHANGES	0
#define SHOW_FURY_FRAMES	0
#define LOG_FRAMES		0
#define SHOW_GENERATION_MAX	0

/* includes */
#include "ipxfury.h"
#include <string.h>
#include <netinet/in.h>

/* timer stuff */
#define FLCP_SET_TIMER(x)	add_event(ifg.events, ETYPE_FLCP_TIMEOUT, 2, 0, (void (*)(void *, int))x->timeout_handler, (void *)x)
#define FLCP_TIMER_OFF(x)	x->restart_counter = -1; remove_events(ifg.events, ETYPE_FLCP_TIMEOUT, (void (*)(void *, int))x->timeout_handler)

/* protos */
int		process_fury_ack(INTERFACE *fdev, BYTE *data, int data_length);

void		send_lcp(INTERFACE *iface, BYTE lcp_code, BYTE id, BYTE *data, int length);
static void	process_lcp(INTERFACE *iface, BYTE *data, int length);

int		process_fury_command(INTERFACE *fdev, BYTE *data, int data_len);


static void	initial_event(INTERFACE *iface, int event);
static void	starting_event(INTERFACE *iface, int event);
static void	closed_event(INTERFACE *iface, int event);
static void	stopped_event(INTERFACE *iface, int event);
static void	closing_event(INTERFACE *iface, int event);
static void	stopping_event(INTERFACE *iface, int event);
static void	req_sent_event(INTERFACE *iface, int event);
static void	ack_rcvd_event(INTERFACE *iface, int event);
static void	ack_sent_event(INTERFACE *iface, int event);
static void	opened_event(INTERFACE *iface, int event);
void		fury_event(INTERFACE *iface, int event);

int		fury_process_frame(INTERFACE *fdev);
int		fury_receive(INTERFACE *fdev, BYTE *data, int length);

int		fury_byte_stuff(BYTE *indata, int length, BYTE *stuffed);

extern GLOBALS	ifg;
extern char *state_str[];

char *event_str[] = {
	"FURY_EVENT_UP",
	"FURY_EVENT_DOWN",
	"FURY_EVENT_OPEN",
	"FURY_EVENT_CLOSE",
	"FURY_EVENT_TON",
	"FURY_EVENT_TOZ",
	"FURY_EVENT_RCR",
	"FURY_EVENT_RCR_BAD",
	"FURY_EVENT_RCA",
	"FURY_EVENT_RCN",
	"FURY_EVENT_RTR",
	"FURY_EVENT_RTA",
};

/****************************************************************************
***********                COMMAND PROCESSORS                     ***********
****************************************************************************/

/*
**	PROCESS_FURY_COMMAND - handles a fury command packet
*/
int process_fury_command(INTERFACE *fdev, BYTE *data, int data_len)
{

	/* 1st byte is the command code */
	switch( *data ) {
		case FCMD_CLEAR_HDRCACHE:
			process_hdrcache_clear(fdev, (data+1), (data_len-1));
			break;
			
		case FCMD_CLEAR_PKTCACHE:
			process_pktcache_clear(fdev, (data+1), (data_len-1));
			break;
			
		case FCMD_RESET:
			/* only do this if we are furious! */
			if( ifg.status == FSTATUS_FURIOUS ) {
				notice("Received reset command!\n");
				command_reset(0);	/* clean everything out but don't send a command */
				/* command reset sets our state to RESETTING */
				send_fury_ack(fdev, FACK_RESET, NULL, 0);	/* send an ack */
			}
			break;
			
		case FCMD_DIFF_RESET:
			/* had a problem with a diff */
			notice("[fury] received diff reset!\n");
			*fdev->cache_ptrs.last_sendsize = -1;
			break;
			
		default:
			errormsg(0, "proc_fury_cmd", "unknown command code (%d)", *data);
			
	}
	
	/* done */
	return(1);

}



/****************************************************************************
***********                ACK PROCESSORS                           *********
****************************************************************************/

/*
**	PROCESS_ACK - handles various types of ACKs
*/
int process_fury_ack(INTERFACE *fdev, BYTE *data, int data_len)
{
	switch( *data ) {
		case FACK_RESET:		/* reset ack */
			/* change the state back to furious */
			/* only do this if we are RESETTING */
			if( ifg.status == FSTATUS_RESETTING ) {
				ifg.status = FSTATUS_FURIOUS;
				notice("Received reset ack.. now returning to furification..\n");
				send_fury_ack(fdev, FACK_RESET, NULL, 0);
			}
			break;
			
		default:
			errormsg(0, "proc_ack", "unknown ACK type (%d)", *data);
			break;
	}
	
	/* done */
	return(1);
}





/****************************************************************************
***********                  SENDING FRAMES                    **************
****************************************************************************/

/*
**	SEND_FURY_COMMAND - send a command their way
*/
int send_fury_command(INTERFACE *fdev, BYTE cmd_type, BYTE *data, int data_len)
{
	char		buffer[CMD_MAX_SIZE+1];
	
	/* did they send too much? */
	if( data_len > CMD_MAX_SIZE ) {
		errormsg(0, "send_fury_cmd", "command data to large (%d bytes)", data_len);
		return(0);
	}
	
	/* 1st byte is the command type */
	buffer[0] = cmd_type;
	
	/* copy the data */
	if( data_len ) {
		memcpy(&buffer[1], data, data_len);
	}
	
	/* ship it out */
	send_fury_frame(fdev, FTYPE_COMMAND, buffer, (data_len+1));
	
	/* done */
	return(1);
}



/*
**	SEND_FURY_ACK - sends an ACK
*/
int send_fury_ack(INTERFACE *fdev, BYTE ack_type, BYTE *data, int data_len)
{
	BYTE		buffer[ACK_MAX_SIZE+1];
	
	/* is it too big? */
	if( data_len >= ACK_MAX_SIZE ) {
		errormsg(0, "send_fury_ack", "ack data to large (%d bytes)", data_len);
		return(0);
	}
	
	/* 1st byte is the ack type */
	buffer[0] = ack_type;
	
	/* copy the data */
	if( data_len ) {
		memcpy(&buffer[1], data, data_len);
	}
	
	/* send it */
	send_fury_frame(fdev, FTYPE_ACK, buffer, (data_len+1));
	
	/* ok */
	return(1);
	
}



/*
**	SEND_FURY_FRAME - send a frame out the given interface
*/
void send_fury_frame(INTERFACE *iface, int type, BYTE *data, int data_len)
{
 	char		buffer[(FURY_FRAME_MAX*2)+1];	/* double for stuffing +1 for end flag */	
	unsigned short	header;
	int		stuffed_len;
	char		d_buffer[256];
	int		d_length;
	DIFF_ARGS	d_args;
	
	
	/* is the outbuf legal? */
	if( iface->outbuf == NULL ) {
		errormsg(0, "send_fury_frame", "no outbuf!");
		return;
	}
	
	/* do not send anything that is NOT LCP before OPENED */
	if( (iface->state != FURY_STATE_OPENED) && (type != FTYPE_LCP) ) {
#if 0	
		errormsg(0, "send_fury_frame", "interface %s is not ready", iface->name);
#endif
		return;
	}

	/* should we send a leading FLAG? -- for serial devices */
	if( iface->frames_sent == 0 ) 	{
		/* put a flag in the buffer */
		buffer_insertb(iface->outbuf, FURY_FLAG);
	}
	
	/* see if we can diff the fury payload..			*/
	/* only do this for IPX packets.. all others need to be clear	*/
	if( (data_len < DIFF_MAX_LENGTH) && ((type & 0x07) == FTYPE_IPX) ) {
		/* try it */
		d_args.inbuf 	= data;
		d_args.insize	= data_len;
		d_args.lastbuf	= iface->cache_ptrs.last_senddata;
		d_args.lastsize	= *iface->cache_ptrs.last_sendsize;
		d_args.outbuf	= d_buffer;
		d_args.outsize	= sizeof(d_buffer);
		
		/* compress */
		d_length = compress_d2(&d_args);
		if( d_length < 0 ) {
			d_length = 0x0FFFFFFF;
		}
	} else {
		/* set bad length */
		d_length = 0x0FFFFFFF;
	}

	/* save this data as the last sent -- the actual frame! */
	/* only if it is IPX data				*/
	if( (type & 0x07) == FTYPE_IPX ) {
		memcpy(iface->cache_ptrs.last_senddata, data, data_len);
		*iface->cache_ptrs.last_sendsize = data_len;
		
		/* check for full first frame! */
		if( (iface->options & FOPT_FULLFRAME) && (buffer_used(iface->outbuf) == 0) ) {
			/* do NOT diff this! */
			/* do nothing.. just stuff the data */
			stuffed_len = fury_byte_stuff(data, data_len, &buffer[2]);
			
			/* reset generation count */
			iface->generation = 0;
			
		} else {
			/* try whatever we can! -- */
			/* what should we use? the diff or nothing? -- or have we reached a max gen */
			/* -+- REMOVE -+- */
			/* d_length = 0x0FFFFFFF; */
			if( (d_length < data_len) && (iface->generation < iface->generation_max) ) {
				/* nope, so diff it */
				/* send the diff -- fix the type */
				type |= FTYPE_DIFF;
	
				/* adjust data_len */
				data_len = d_length;
			
				/* stuff the data */
				stuffed_len = fury_byte_stuff(d_buffer, d_length, &buffer[2]);
				
				/* up the generation counter */
				iface->generation++;
			} else {
			
#if SHOW_GENERATION_MAX == 1
				if( iface->generation >= iface->generation_max ) {
					notice("Reached Generation max!\n");
				}
#endif
				/* do nothing.. just stuff the data */
				stuffed_len = fury_byte_stuff(data, data_len, &buffer[2]);
				
				/* reset generation count */
				iface->generation = 0;
			}
		}
	} else {
		/* non-IPX.. don't do anything special */
		stuffed_len = fury_byte_stuff(data, data_len, &buffer[2]);
	}
		
	/* put in the trailing FLAG (+2 for the fury header) */
	buffer[stuffed_len + 2] = FURY_FLAG;
		
	/* now create the header */
	header = (type & 0xC000) | ((type & 0x0007) << 11) | (data_len + 2);
	header = htons(header);		/* fix byte order */
	*((unsigned short *)&buffer[0]) = header;

	
#if LOG_FRAMES == 1
	switch( type & 0x07 ) {
		case FTYPE_IPX:
			fprintf(ifg.logfile, "Sending IPX Data:\n");
			fhexdump(ifg.logfile, buffer, stuffed_len+3);
			break;
		case FTYPE_ACK:
			fprintf(ifg.logfile, "Sending ACK:\n");
			fhexdump(ifg.logfile, buffer, stuffed_len+3);
			break;
		case FTYPE_COMMAND:
			fprintf(ifg.logfile, "Sending CMD:\n");
			fhexdump(ifg.logfile, buffer, stuffed_len+3);
			break;
		default:
			fprintf(ifg.logfile, "Sending UNKNOWN!:\n");
			fhexdump(ifg.logfile, buffer, stuffed_len+3);
			break;
	}
#endif	
	/* send the buffer out */
#if 0
	notice("[fury] Fury frame size: %d bytes\n", stuffed_len+3);
#endif
	buffer_insert(iface->outbuf, buffer, stuffed_len+3);
	
	/* inc frames sent */
	iface->frames_sent++;
	
	/* done */
	return;

}



/****************************************************************************
*************        FURY FRAME READERS/PROCESSORS                ***********
****************************************************************************/

/*
**	FURY_PROCESS_FRAME - processes a fury frame
**
*/
int fury_process_frame(INTERFACE *fdev)
{
	BYTE			*frame_data;
	unsigned short		header;
	int			data_length;
	FURYFRAME		*fury;
	char			d_buffer[256];
	int			d_length;
	DIFF_ARGS		d_args;
	
	/* set fury ptr */
	fury = &fdev->furyframe;
	
	/* is there anything to process? */
	if( fury->count == 0 ) {
		return(0);
	}
	
	/* convert the header into host byte order */
	header = *((unsigned short *)fury->data);	/* header is NBO */
	header = ntohs(header);				/* now HBO :) */

	/* do the numbers match? */
	if( fury->count != FHDR_LENGTH(header) ) {
		/* slight problem here! */
		errormsg(0, "fury_process", "buffer length (%d) does not match header (%d)", fury->count, FHDR_LENGTH(header));
		hexdump(fury->data, fury->count);
		
		/* failed */
		return(0);
	}
	
	/* set the data length (-2 for the 2 byte header) */
	data_length = (fury->count - 2);
	frame_data = &fury->data[2];
	
	/* is this a DIFF? */
	if( header & FTYPE_DIFF ) {
		/* yes.. so we need to make it! */
		d_args.inbuf	= frame_data;
		d_args.insize	= data_length;
		d_args.lastbuf	= fdev->cache_ptrs.last_recvdata;
		d_args.lastsize	= *fdev->cache_ptrs.last_recvsize;
		d_args.outbuf	= d_buffer;
		d_args.outsize	= sizeof(d_buffer);
		
		/* ok, undiff */
		d_length = uncompress_d2(&d_args);
		if( d_length < 0 ) {
			/* need to send something back.. if we haven't in awhile */
			if( *fdev->cache_ptrs.diff_reset <= 0 ) {
				errormsg(0, "fury_proc", "error undiffing frame! (reset sent!)");
				/* set the starting counter */
				*fdev->cache_ptrs.diff_reset = FURY_DIFF_RESET_START;
				/* send the command */
				send_fury_command(fdev, FCMD_DIFF_RESET, NULL, 0);
			} else {
				/* dec the counter */
				*fdev->cache_ptrs.diff_reset = *fdev->cache_ptrs.diff_reset - 1;
			}
			
			/* bad last recv frame */
			*fdev->cache_ptrs.last_recvsize = -1;
			
			/* fail */
			return(0);
		}
		
		/* adjust ptrs */
		frame_data 	= d_buffer;
		data_length	= d_length;
	}
	
	/* save this as the last received frame */
	/* if it is an IPX frame		*/
	if( FHDR_TYPE(header) == FTYPE_IPX ) {
		memcpy(fdev->cache_ptrs.last_recvdata, frame_data, data_length);
		*fdev->cache_ptrs.last_recvsize = data_length;
	}
		
	/* ok, everything seems ok, lets process */
	switch( FHDR_TYPE(header) ) {
		case FTYPE_LCP:
			process_lcp(fdev, frame_data, data_length);
			break;
		
		case FTYPE_IPX:
			if( ! process_ipx(fdev, header, frame_data, data_length) ) {
				/* problem with this data.. call error handler */
			}
			break;
			
		case FTYPE_ACK:
			process_fury_ack(fdev, frame_data, data_length);
			break;
			
		case FTYPE_COMMAND:
			process_fury_command(fdev, frame_data, data_length);
			break;
			
		default:
			errormsg(0, "fury_process", "unknown type (%d)", FHDR_TYPE(header));
			break;
	}
	
	/* done */
	return(1);
}



/*
**	FURY_RECEIVE - called when there are some bytes to handle
**
**	Note: This routine unstuffs the incoming data (moron!)
*/
int fury_receive(INTERFACE *fdev, BYTE *data, int length)
{
	FURYFRAME	*fury;
	BYTE		ch, *buf;

	/* set the pointer */
	fury = &fdev->furyframe;

	/* position the buffer pointer for writing */
	buf = &fury->data[fury->count];

	/* while there are still bytes to receive */
	while( length-- > 0  ) {
		/* get a byte */
		ch = *data++;
		
		/* based on the byte, do the following.. */
		switch( ch ) {
			case FURY_FLAG:	
				/* process this frame if drop is not set */
				if( (fury->drop == 0x00) && (fury->count != 0) ) {
#if SHOW_FURY_FRAMES == 1				
					notice("Rcvd FURY Frame:\n");
					hexdump(fury->data, fury->count);
#endif
					fury_process_frame(fdev);
				}
				fury->drop = 0x00;	/* stop dropping */
				
				/* reset the data to start over */
				fury->count = 0;
				fury->escape = 0x00;
				buf = fury->data;
							
				break;
				
			case FURY_ESCAPE:
				/* mess with the next byte */
				fury->escape = FURY_INVERT;
				break;
				
			default:
				/* stick the byte into the buffer, if we have room */
				/* we check against MAX*2 because of byte stuffing */
				if( fury->count < (FURY_FRAME_MAX*2) ) {
					*buf++ = ch ^ fury->escape;
					fury->escape = 0x00;
					fury->count++;
					break;
				} else {
					errormsg(0, "fury_recv", "frame was to large (discarding..)");
					fury->count = 0;
					fury->drop = 1;		/* ignore till next FLAG */
				}
				break;
		}
	
	} /* while length */
	
	/* done */
	return(1);
}





/****************************************************************************
***************            LCP HANDLING FUNCTIONS                  **********
****************************************************************************/

/*
**	SEND_LCP - sends an LCP packet
*/
void send_lcp(INTERFACE *iface, BYTE lcp_code, BYTE id, BYTE *data, int data_len)
{
	BYTE			buffer[FLCP_PACKET_MAX];
	FLCP_PACKETHEADER	*header;
	
	/* truncate length */
	if( data_len > FLCP_DATA_MAX ) {
		errormsg(0, "send_lcp", "truncated data %d -> %d", data_len, FLCP_DATA_MAX);
		data_len = FLCP_DATA_MAX;
	}

	/* construct the packet header */
	header = (FLCP_PACKETHEADER *)buffer;
	header->code = lcp_code;
	header->id = id;
	header->length = htons(data_len+4);
	
	/* copy the data */
	if( data_len > 0 ) {
		memcpy(&buffer[4], data, data_len);
	}
	
	/* send this packet out (+4 for header) */
	send_fury_frame(iface, FTYPE_LCP, buffer, (data_len + sizeof(FLCP_PACKETHEADER)));	
		
	/* done */
}


/*
**	PROCESS_LCP - handles link control packets
*/
static void process_lcp(INTERFACE *iface, BYTE *data, int length)
{
	int		lcp_length;
	struct timeval	*then, now;
	
	/* must be a valid LCP packet, (length >= 4) */
	if( length < 4 ) {
		errormsg(0, "process_lcp", "lcp packet to small (%d bytes)", length);
		return;
	}
	
	/* make sure the lengths agree */
	lcp_length = *((unsigned short *)&data[2]);
	lcp_length = ntohs(lcp_length);
	if( length != lcp_length ) {
		errormsg(0, "process_lcp", "lengths do not agree (%d/%d)", length, lcp_length);
		return;
	}		
	
	/* should we process this? */
	if( iface->flags & FURY_IF_LCP_IGNORE ) {
		errormsg(0, "process_lcp", "ignoring LCP packet");
		return;
	}
	
	/* now check the code */
	switch( data[0] ) {
		case FLCP_CONFIG_REQ:
			iface->event_handler(iface, FURY_EVENT_RCR);
			break;
		case FLCP_CONFIG_ACK:
			iface->event_handler(iface, FURY_EVENT_RCA);
			break;
		case FLCP_CONFIG_NAK:	
			iface->event_handler(iface, FURY_EVENT_RCN);
			break;
		case FLCP_CONFIG_REJ:	/* don't have one */
			break;
		case FLCP_TERMINATE_REQ:
			iface->event_handler(iface, FURY_EVENT_RTR);
			break;
		case FLCP_TERMINATE_ACK:
			iface->event_handler(iface, FURY_EVENT_RTA);
			break;
		
		case FLCP_ECHO_REQ:
			if( iface->state == FURY_STATE_OPENED ) {
				send_lcp(iface, FLCP_ECHO_REPLY, 0x00, &data[4], length-4);
			}
			break;
			
		case FLCP_ECHO_REPLY:
			if( iface->state == FURY_STATE_OPENED ) {
				gettimeofday(&now, NULL);
 				then = (struct timeval *)&data[4];
 				now.tv_usec += (now.tv_sec - then->tv_sec) * 1000000;
 				
 				notice("Ping (%dms)\n", (int)((now.tv_usec - then->tv_usec)/1000));
			}
			break;
			
		default:
			errormsg(0, "process_lcp", "bad lcp packet type (%d)", data[0]);
			break;	
	}


}





/*************************************************************************
***************      EVENT HANDLERS (for the automaton)             ******
*************************************************************************/


/*
**	INITIAL_EVENT (0) - handles events in the initial state
*/
static void initial_event(INTERFACE *iface, int event)
{
	switch( event ) {
		case FURY_EVENT_OPEN:		/* lower level ready */
			iface->state = FURY_STATE_STARTING;
			/* tls */
			break;
		
		case FURY_EVENT_UP:		/* dunno if this will happen */
			iface->state = FURY_STATE_CLOSED;
			break;
			
		/* anything else, just stay here */
	}

}


/*
**	STARTING_EVENT (1) - handle events in the starting state
*/
static void starting_event(INTERFACE *iface, int event)
{
	switch( event ) {
		case FURY_EVENT_UP:	/* carrier detected, or so */
			/* irc */
			iface->restart_counter = (FURY_RESTART_CONFIGURE_MAX - 1);
			FLCP_SET_TIMER(iface);
			/* scr */
			send_lcp(iface, FLCP_CONFIG_REQ, 0xab, NULL, 0);
			/* move to new state */
			iface->state = FURY_STATE_REQ_SENT;
			notice("[%6s] Establishing connection...\n", iface->name);
			break;
	}

}



/*
**	CLOSED_EVENT (2) -
*/
static void closed_event(INTERFACE *iface, int event)
{
	switch( event ) {
		case FURY_EVENT_OPEN:
			/* irc */
			iface->restart_counter = (FURY_RESTART_CONFIGURE_MAX-1);
			FLCP_SET_TIMER(iface);
			/* scr */
			send_lcp(iface, FLCP_CONFIG_REQ, 0x40, NULL, 0);
			/* new state */
			iface->state = FURY_STATE_REQ_SENT;
			break;
			
		case FURY_EVENT_DOWN:
			iface->state = FURY_STATE_INITIAL;
			break;
			
		case FURY_EVENT_RCR:
		case FURY_EVENT_RTR:
			/* sta */
			send_lcp(iface, FLCP_TERMINATE_ACK, 0xc0, NULL, 0);
			break;
			
	}
}



/*
**	STOPPED_EVENT (3) -
*/
static void stopped_event(INTERFACE *iface, int event)
{
	switch( event ) {
		case FURY_EVENT_RCR:
			/* irc */
			iface->restart_counter = (FURY_RESTART_CONFIGURE_MAX-1);
			FLCP_SET_TIMER(iface);
			/* scr */
			send_lcp(iface, FLCP_CONFIG_REQ, 0x30, NULL, 0);
			/* sca */
			send_lcp(iface, FLCP_CONFIG_ACK, 0x31, NULL, 0);
			/* new state */
			iface->state = FURY_STATE_ACK_SENT;
			break;
			
		case FURY_EVENT_CLOSE:
			/* new state */
			iface->state = FURY_STATE_CLOSED;
			break;
			
		
		case FURY_EVENT_RTR:
		case FURY_EVENT_RCA:
			/* sta */
			send_lcp(iface, FLCP_TERMINATE_ACK, 0xf0, NULL, 0);
			break;
	}
}



/*
**	CLOSING_EVENT (4) -
*/
static void closing_event(INTERFACE *iface, int event)
{
	switch( event ) {
		case FURY_EVENT_TON:		/* timeout, count > 0 */
			/* dec counter */
			iface->restart_counter--;
			/* str */
			send_lcp(iface, FLCP_TERMINATE_REQ, 0x20, NULL, 0);
			break;
			
		case FURY_EVENT_TOZ:		/* timeout, count = 0 */
			/* timer off */
			FLCP_TIMER_OFF(iface);
			/* new state */
			iface->state = FURY_STATE_CLOSED;
			break;
			
		case FURY_EVENT_RTR:
			/* sta */
			send_lcp(iface, FLCP_TERMINATE_ACK, 0xb0, NULL, 0);
			break;
			
		case FURY_EVENT_RTA:
			/* timer off */
			FLCP_TIMER_OFF(iface);
			/* new state */
			iface->state = FURY_STATE_CLOSED;
			break;
			
		case FURY_EVENT_OPEN:
			/* remove timer */
			FLCP_TIMER_OFF(iface);
			/* new state */
			iface->state = FURY_STATE_STOPPING;
			break;
	}
}



/*
**	STOPPING_EVENT (5) -
*/
static void stopping_event(INTERFACE *iface, int event)
{
	switch( event ) {
		case FURY_EVENT_TON:
			/* dec the counter */
			iface->restart_counter--;
			/* str */
			send_lcp(iface, FLCP_TERMINATE_REQ, 0x10, NULL, 0);		
			break;
			
			
		case FURY_EVENT_RTR:
			/* sta */
			send_lcp(iface, FLCP_TERMINATE_ACK, 0x11, NULL, 0);
			break;
			
		case FURY_EVENT_TOZ:		/* zero counter */
		case FURY_EVENT_RTA:		/* term ack */
			/* stop timer */
			FLCP_TIMER_OFF(iface);
			/* new state */
			iface->state = FURY_STATE_STOPPED;
			break;
	}
}



/*
**	REQ_SENT_EVENT (6) -
*/
void req_sent_event(INTERFACE *iface, int event)
{
	switch( event ) {
		case FURY_EVENT_RCA:	/* rcvd config ack */
			/* irc */
			iface->restart_counter = (FURY_RESTART_CONFIGURE_MAX-1);
			FLCP_SET_TIMER(iface);
			/* move to next state */
			iface->state = FURY_STATE_ACK_RCVD;
			break;
			
		case FURY_EVENT_RCR:	/* rcvd config req */
			/* sca */
			send_lcp(iface, FLCP_CONFIG_ACK, 0x55, NULL, 0);
			/* new state */
			iface->state = FURY_STATE_ACK_SENT;
			break;
		
		case FURY_EVENT_TOZ:	/* timeout, counter == 0 */
			/* stop the timer */
			FLCP_TIMER_OFF(iface);
			/* new state */
			notice("[%6s] Failed to establish connection!\n", iface->name);
			iface->state = FURY_STATE_STOPPED;
			break;
			
		case FURY_EVENT_TON:	/* timeout, count != 0 */
			/* reset the timer */
			iface->restart_counter--;	/* dec counter */
			/* scr */
			send_lcp(iface, FLCP_CONFIG_REQ, 0x64, NULL, 0);
			break;
			
		case FURY_EVENT_RTR:	/* rcvd terminate req */
			/* sta */
			send_lcp(iface, FLCP_TERMINATE_ACK, 0x75, NULL, 0);
			break;
			
		case FURY_EVENT_CLOSE:
			/* irc */
			iface->restart_counter = (FURY_RESTART_TERMINATE_MAX-1);
			FLCP_SET_TIMER(iface);
			/* new state */
			iface->state = FURY_STATE_CLOSING;
			break;
			
		case FURY_EVENT_DOWN:	/* carrier loss */
			iface->state = FURY_STATE_STARTING;
			break;
			
	}

}


/*
**	ACK_RCVD_EVENT (7) -
*/
static void ack_rcvd_event(INTERFACE *iface, int event)
{
	switch( event ) {
		case FURY_EVENT_TON:	/* timeout, count > 0 */
			/* dec count */
			iface->restart_counter--;
			/* scr */
			send_lcp(iface, FLCP_CONFIG_REQ, 0xb0, NULL, 0);
			/* new state */
			iface->state = FURY_STATE_REQ_SENT;
			break;
			
		case FURY_EVENT_TOZ:	/* timeout, count = 0 */
			/* timer off */
			FLCP_TIMER_OFF(iface);
			/* new state */
			iface->state = FURY_STATE_STOPPED;
			break;
			
		case FURY_EVENT_RCR:	/* rcvd config req */
			/* sca */
			send_lcp(iface, FLCP_CONFIG_ACK, 0x12, NULL, 0);
			/* turn timer off */
			FLCP_TIMER_OFF(iface);
			/* new state */
			iface->state = FURY_STATE_OPENED;
			notice("[%6s] Connection opened.\n", iface->name);
			/* ignore the next few LCP packets for the next 3 seconds */
			iface->flags |= FURY_IF_LCP_IGNORE;
			add_event(ifg.events, ETYPE_LCP_IGNORE, 3, 0, NULL, (void *)iface);
			break;
			
	}
}


/*
**	ACK_SENT_EVENT (8) -
*/
static void ack_sent_event(INTERFACE *iface, int event)
{
	switch( event ) {
		case FURY_EVENT_TON:	/* timeout, count > 0 */
			/* dec ounter */
			iface->restart_counter--;
			/* scr */
			send_lcp(iface, FLCP_CONFIG_REQ, 0xa0, NULL, 0);
			break;
			
		case FURY_EVENT_TOZ:	/* timeout, count = 0 */
			/* timer off */
			FLCP_TIMER_OFF(iface);
			/* new state */
			iface->state = FURY_STATE_STOPPED;
			break;
			
		case FURY_EVENT_RCA:	/* rcvd config ack */
			/* timer off */
			FLCP_TIMER_OFF(iface);
			/* new state */
			iface->state = FURY_STATE_OPENED;
			notice("[%6s] Connection opened.\n", iface->name);
			/* ignore the next few LCP packets for the next 3 seconds */
			iface->flags |= FURY_IF_LCP_IGNORE;
			add_event(ifg.events, ETYPE_LCP_IGNORE, 3, 0, NULL, (void *)iface);
			break;
	}

}



/*
**	OPENED_EVENT (9) -
*/
static void opened_event(INTERFACE *iface, int event)
{
	switch( event ) {
		case FURY_EVENT_DOWN:
			/* timer off */
			FLCP_TIMER_OFF(iface);
			/* new state */
			iface->state = FURY_STATE_STARTING;
			notice("[%6s] Connection ended.\n", iface->name);
			break;
		
		case FURY_EVENT_CLOSE:
			/* irc */
			iface->restart_counter = (FURY_RESTART_TERMINATE_MAX-1);
			FLCP_SET_TIMER(iface);
			/* str */
			send_lcp(iface, FLCP_TERMINATE_REQ, 0x50, NULL, 0);
			/* new state */
			iface->state = FURY_STATE_CLOSING;
			notice("[%6s] Closed.\n", iface->name);
			break;
		
		case FURY_EVENT_RCR:
			/* scr */
			send_lcp(iface, FLCP_CONFIG_REQ, 0x51, NULL, 0);
			/* sca */
			send_lcp(iface, FLCP_CONFIG_ACK, 0x52, NULL, 0);
			/* timer */
			iface->restart_counter = (FURY_RESTART_CONFIGURE_MAX-1);
			FLCP_SET_TIMER(iface);
			/* new state */
			iface->state = FURY_STATE_ACK_SENT;
			notice("[%6s] Reconfiguring connection...\n", iface->name);
			break;
			
		case FURY_EVENT_RTR:
			/* says zrc */
			iface->restart_counter = 0;
			FLCP_SET_TIMER(iface);
			/* sta */
			send_lcp(iface, FLCP_TERMINATE_ACK, 0x53, NULL, 0);
			/* new state */
			iface->state = FURY_STATE_STOPPING;
			notice("[%6s] Termination request received.\n", iface->name);
			break;
			
	}
}



/*
**	FURY_EVENT - moves the automaton around
*/
void fury_event(INTERFACE *iface, int event)
{

#if SHOW_STATE_CHANGES == 1
	int		oldstate;
	oldstate = iface->state;
#endif

	/* different events effect different states, so find out the state */
	switch( iface->state ) {
		case FURY_STATE_INITIAL:	/* this is where we start */
			initial_event(iface, event);
			break;
		case FURY_STATE_STARTING:
			starting_event(iface, event);			
			break;
		case FURY_STATE_CLOSED:
			closed_event(iface, event);
			break;
		case FURY_STATE_STOPPED:
			stopped_event(iface, event);
			break;
		case FURY_STATE_CLOSING:
			closing_event(iface, event);
			break;
		case FURY_STATE_STOPPING:
			stopping_event(iface, event);
			break;
		case FURY_STATE_REQ_SENT:
			req_sent_event(iface, event);
			break;
		case FURY_STATE_ACK_RCVD:
			ack_rcvd_event(iface, event);
			break;
		case FURY_STATE_ACK_SENT:
			ack_sent_event(iface, event);
			break;
		case FURY_STATE_OPENED:
			opened_event(iface, event);
			break;
	} /* switch on state */

#if SHOW_STATE_CHANGES == 1
	notice("State Change: %s -- %s --> %s\n", state_str[oldstate], event_str[event], state_str[iface->state]);
#endif

}







/*****************************************************************************
*****************           MISC ROUTINES                       **************
*****************************************************************************/

/*
**	FURY_STUFF
*/
int fury_byte_stuff(BYTE *indata, int length, BYTE *stuffed)
{
	int		newlength;
	BYTE		escape;
	int		oldlength;
	
	oldlength = length;
	
	/* stuff the data */
	newlength = 0;
	escape = 0;
	
	while( length ) {
		if( (*indata == FURY_FLAG) || (*indata == FURY_ESCAPE) ) {
			*stuffed++ = FURY_ESCAPE;	/* insert escape */
			newlength++;			/* inc length */
			escape = FURY_INVERT;		/* set xor */
		} 
		*stuffed++ = *indata++ ^ escape;
		escape = 0x00;
		
		/* 1 less data byte to deal with */
		length--;
		
		/* add to the stuffed size */
		newlength++;
	}

	/* return new length */
	return(newlength);

}


